/****************************************************************************
 *
 *   File: finch_interface.cc:
 *   
 *   Author: Mark Stephenson
 *   Date: Tue Jun  3 17:27:15 2003
 *
 *   Function: The interface to finch's library.
 *
 ***************************************************************************/


#include "finch_interface.h"
#include "configure_expr.h"
#include <stdio.h>

static genetic_expr     *FINCH_curr_expr = NULL;
static genetic_expr_bnf *FINCH_curr_bnf  = NULL;

void
FINCH_initialize_lib ()
///////////////////////////////////////////////////////////////
//
// Initialize the finch library.  This function should
// only be called once-- when your pass initializes.
// 
///////////////////////////////////////////////////////////////
{
  // Get the FINCHHOME environment variable.  This is 
  // the base directory that we will use for all 
  // communication with the harness.
  char* finch_home = getenv ("FINCHHOME");
  
  finch_claim ((finch_home != NULL), 
	       "Please set the FINCHHOME environment variable.");
  
  finch_claim ((FINCH_curr_expr == NULL),
	       "Please only call FINCH_initialize_lib once.");
  
  // If the config file is not present, we have to create it.
  string config_fname = string(finch_home) + string("/config");
  string pfun_fname = string(finch_home) + string("/evaluate/pfun.l");

  // Otherwise, go ahead and create the genetic expressions.
  ifstream config_file (config_fname.c_str());
  FINCH_curr_bnf = (config_file) ? 
    new genetic_expr_bnf_default (config_file) : NULL;
    
  if (FINCH_curr_bnf)
  {
    FINCH_curr_expr = genetic_expr::parse_expr (FINCH_curr_bnf, 
						pfun_fname.c_str());
    finch_claim (FINCH_curr_expr, "Cannot open the priority function.");
  }
  else
  {
    FINCH_curr_expr = new configure_expr (config_fname);
  }
}


///////////////////////////////////////////////////////////////
//
// Evaluate the current priority function with the arguments
// that are passed in.
//
///////////////////////////////////////////////////////////////
double 
FINCH_evaluate_real_expr (unsigned num_dbls, unsigned num_bools, ...)
{
  va_list ap;
  double *dlist;
  int32  *blist;

  dlist = new double[num_dbls];
  blist = new int32[num_bools];

  va_start (ap, num_bools);
  for (uint32 i = 0; i < num_dbls; i++)
    dlist[i] = va_arg(ap, double);
  for (uint32 j = 0; j < num_bools; j++)
    blist[j] = va_arg(ap, int32);
  va_end (ap);

  lval val = FINCH_curr_expr->evaluate (dlist, blist, 
					num_dbls, num_bools);

  delete[] dlist;
  delete[] blist;
  return val.d;
}

int 
FINCH_evaluate_binary_expr (unsigned num_dbls, unsigned num_bools, ...)
{
  va_list ap;
  double *dlist;
  int32  *blist;

  dlist = new double[num_dbls];
  blist = new int32[num_bools];

  va_start (ap, num_bools);
  for (uint32 i = 0; i < num_dbls; i++)
    dlist[i] = va_arg(ap, double);
  for (uint32 j = 0; j < num_bools; j++)
    blist[j] = va_arg(ap, int32);
  va_end (ap);

  lval val = FINCH_curr_expr->evaluate (dlist, blist, 
					num_dbls, num_bools);

  delete[] dlist;
  delete[] blist;
  return val.i;
}

double 
FINCH_evaluate_real_array (unsigned num_dbls, unsigned num_bools,
                           double *dbls, int *bools)
{
  lval val = FINCH_curr_expr->evaluate (dbls, bools, 
					num_dbls, num_bools);
  return val.d;
}

int
FINCH_evaluate_binary_array (unsigned num_dbls, unsigned num_bools,
                             double *dbls, int *bools)
{
  lval val = FINCH_curr_expr->evaluate (dbls, bools, 
					num_dbls, num_bools);
  return val.i;
}

///////////////////////////////////////////////////////////////
//
// Delete any dynamically allocated memory that we have
// created.
//
///////////////////////////////////////////////////////////////
void   
FINCH_finalize_lib ()
{
  delete FINCH_curr_expr;
  delete FINCH_curr_bnf;
}

