/******************************************************************
 *
 *   File: genetic_expr.h:
 *   
 *   Author: Mark Stephenson
 *   Date: Mon Jul 22 14:10:49 2002
 *
 *   Function: A base class for genetic expressions (sexprs).
 *
 *****************************************************************/

#ifndef __GENETIC_EXPR_H__
#define __GENETIC_EXPR_H__


#include "genetic_expr_bnf.h"
#include "genetic_lexer.h"  
#include <fstream>
#include <string>
#include <vector>


///////////////////////////////////////////
//////////// genetic expression ///////////

class genetic_expr
{
 public:

  // Creates a random expression that is probably of
  // the specified height.
  static genetic_expr* random (genetic_expr_bnf *bnf,
			       uint32 height,
			       genetic_expr* p = NULL,
			       uint32 type = DOUBLE);

  // Parses the expression from a file.
  static genetic_expr* parse_expr (genetic_expr_bnf*, istream&);
  static genetic_expr* parse_expr (genetic_expr_bnf*, const char*);

  // Simplifies the current expression.
  virtual genetic_expr* simplify ()  { return this->clone (); }

  // Functions for printing the expression.
  virtual void print (ostream& os, bool = false, bool = false) const;

  // Returns true if the two expressions are equal.
  virtual bool operator == (const genetic_expr&) const;

  // Evaluates the expression in the 
  // current environment.
  virtual lval evaluate (double[], int32[], uint32, uint32);
  virtual lval evaluate (double[], int32[]) = 0;

  // Cross this expression with the one that's passed in.
  virtual void cross (genetic_expr*, genetic_expr*&, genetic_expr*&);

  // Mutates the current expression.
  virtual void mutate ();

  // Mutates the current expression according to Una-May.
  virtual genetic_expr* hvl_mutate ();
  
  // Clones the current expression.
  virtual genetic_expr* clone () = 0;

  // Sets an annotation. This data is not automatically deleted.
  void set_annotation (void *p)    { m_annotation = p; }

  // Returns the annotation.
  void *get_annotation ()          { return m_annotation; }
  
  // Sets the specified fitness element.
  void set_fitness_element (uint32 id, double fitness);

  // Gets the fitness element.
  double get_fitness_element (uint32 id) const;

  // Returns true if a fitness element has been set.
  bool memoized (uint32 id) const;

  // Sets the overall fitness.
  void set_fitness (double f)       { m_total_fitness = f; }

  // Gets the overall fitness.
  double get_fitness () const       { return m_total_fitness; }

  // Returns the lifetime of this expression.
  uint32 lifetime () const          { return m_lifetime; }

  // Increments the lifetime of this expression.
  void inc_lifetime ()              
    { m_lifetime++; if (m_expiry) m_expiry--; }

  // Marks an expiration date on the best expression.
  void mark_as_best (uint32 sunset)  { m_expiry = sunset; }

  // Returns true if this expr was the best expression
  // within sunset gens ago; This works well with DSS.
  bool was_best () const 
    { return (m_expiry); }

  // Returns the number of nodes in this expression.
  uint32 size () const;

  // Deletes the expression.
  virtual ~genetic_expr ();

 protected:

  genetic_expr (genetic_expr_bnf*, genetic_expr*, uint32);

  // Returns the typing of expression this is.
  virtual uint32 typing () const = 0;

  // Returns the opcode.
  virtual char* opcode () const = 0;

  // Returns true iff this expression is a leaf.
  virtual bool is_leaf () const     { return !(this->num_rands()); }

  // Returns true iff this expression is a constant.
  virtual bool is_constant () const { return false; }

  // Returns the height.
  uint32 height () const;

  // Mutates the expression node.
  virtual void mutate_node ();

  // Prints some of the expression's state if bool is true.
  void print_prologue (ostream&, bool, bool) const;

  // Prints some of the expression's state if bool is false.
  void print_epilogue (ostream&, bool) const;

  // Prints the range that the variable took on.
  virtual void print_range (ostream&) const {}

  // Supplants the first expression with the second expression.
  void supplant (genetic_expr*, genetic_expr*);

  // Returns the number of operands in this expression.
  uint32 num_rands () const       { return m_num_rands; }

  // Returns the array of operands.
  genetic_expr** rands () const     { return m_rands; }

  // Returns the parent.
  genetic_expr* get_parent () const { return m_parent; }

  // Sets the parent.
  void set_parent (genetic_expr *p) { m_parent = p; }

  // Copies the arguments operands to this expression.
  void clone_rands (genetic_expr*);

  // Returns a random node from the expression.
  virtual genetic_expr* random_subexpr ();

  // Parses the expression, then makes sure it's return is boolean.
  static genetic_expr *parse_bool_expr (genetic_expr_bnf*, 
					genetic_expr*, genetic_lexer*);

  // Parses the expression, then makes sure it's return is double.
  static genetic_expr *parse_dbl_expr (genetic_expr_bnf*, 
				       genetic_expr*, genetic_lexer*);

  // Finds two subtrees to crossover.
  bool find_subtrees (genetic_expr*, genetic_expr*, 
		      genetic_expr*&, genetic_expr*&) const;

 private:

  // Chooses a random node.
  void choose_random_node (bool, double*, genetic_expr**);

  // Uses the lexer to parse the expression.
  static genetic_expr *parse_expr (genetic_expr_bnf*, 
				   genetic_expr*, genetic_lexer*);

 protected:

  // Information that is common to a "class" of expressions.
  genetic_expr_bnf* m_bnf;

  genetic_expr* m_parent;
  uint32 m_num_rands;
  genetic_expr** m_rands;
  uint32 m_lifetime;
  void *m_annotation;
  uint32 m_expiry;
  double m_total_fitness;
  vector<double> m_fitness_vector;
};


extern void finch_claim (bool test, char *message);


#endif
