/************************************************************************
 *
 *   File: genetic_expr_bnf.cc:
 *   
 *   Author: Mark Stephenson
 *   Date: Sat Mar 22 11:25:26 2003
 *
 *   Function: This file contains classes that describe the bnf of the
 *      expression language.  We provide a "default" bnf that can
 *      be used to create expressions.  However, if you want to 
 *      extend the bnf, you can either extend the genetic_expr_bnf
 *      class, or create a subclass of it.  
 *
 *      Here are directions for extending the default bnf:
 *      
 *      1) Choose a unique opcode for the new function, and add it 
 *         to the list of opcodes in opcodes.h, opcodes.cc.
 *
 *      2) Create a class that derives from double_expr or 
 *         boolean_expr depending on what type your function returns.
 * 
 *      3) Within the new class, create a static function that 
 *         randomly creates said function.
 *
 *      4) Likewise, create a static function for parsing.
 *
 *      For steps 2, 3 and 4, see one of the classes in double_expr.h 
 *      or boolean_expr.h.
 *
 *      5) Add the static random initialization function created in 
 *         step 3 to the appropriate dispatch table.
 *
 *           +-DISPATCH TABLE -+-DESCRIPTION-----------------+
 *           | m_dbl_terms     | Returns double/is terminal. |
 *           | m_dbl_nonterms  | Returns double/is non term. |
 *           | m_bool_terms    | Returns bool/is terminal.   |
 *           | m_bool_nonterms | Returns bool/is non term.   |
 *           +-----------------+-----------------------------+
 *
 *      6) Add the static parsing function created in step 4 to the
 *         genetic_expr_bnf::parse_table.  The entry in the table is
 *         a binding from the opcode you choose in step 1 to the
 *         function you created in step 4.  Look at the table, you'll
 *         see what I'm talking about.
 *
 *      If you wish to subclass genetic_expr_bnf, you shoud know that
 *      this class is simply used a dispatch table for parsing and
 *      creaing random expressions.
 *      
 *
 ************************************************************************/

#include "genetic_expr_bnf.h"
#include "genetic_expr.h"
#include "double_expr.h"
#include "boolean_expr.h"


///// Functions that create random non-terminal DOUBLE-typed exprs //////
expr_nonterm_fun genetic_expr_bnf_default::m_dbl_nonterms[] = {
  &expr_add::random,
  &expr_sub::random,
  &expr_mul::random,
  &expr_cmul::random,
  &expr_div::random,
  &expr_sqrt::random,
  &expr_ternary::random
};

uint32 
genetic_expr_bnf_default::num_dbl_nonterms () const
//////////////////////////////////////////////////////////////
//
//  Returns the number of double-typed non-terminals.
//
//////////////////////////////////////////////////////////////
{
  return sizeof(m_dbl_nonterms) / sizeof(expr_nonterm_fun);
}


////////// Functions that create random DOUBLE-typed terminals ///////
expr_term_fun genetic_expr_bnf_default::m_dbl_terms[] = {
  &expr_dbl_const::random,
  &expr_dbl_arg::random
};

uint32 
genetic_expr_bnf_default::num_dbl_terms () const
//////////////////////////////////////////////////////////////
//
//  Returns the number of double-typed terminal functions.
//
//////////////////////////////////////////////////////////////
{
  return sizeof(m_dbl_terms) / sizeof(expr_term_fun);
}


///////  Functions that create random non-terminal BOOLEANs ////////
expr_nonterm_fun genetic_expr_bnf_default::m_bool_nonterms[] = {
  &expr_lt::random,
  &expr_gt::random,
  &expr_eq::random,
  &expr_and::random,
  &expr_or::random,
  &expr_not::random
};

uint32 
genetic_expr_bnf_default::num_bool_nonterms () const
//////////////////////////////////////////////////////////////
//
//  Returns the number of Boolean non-terminal functions.
//
//////////////////////////////////////////////////////////////
{
  return sizeof(m_bool_nonterms) / sizeof(expr_nonterm_fun);
}

//// Functions that create random terminal BOOLEAN expressions /////
expr_term_fun genetic_expr_bnf_default::m_bool_terms[] = {
  &expr_bool_const::random,
  &expr_bool_arg::random
};

uint32 
genetic_expr_bnf_default::num_bool_terms () const
//////////////////////////////////////////////////////////////
//
//  Returns the number of Boolean terminal functions.
//
//////////////////////////////////////////////////////////////
{
  return sizeof(m_bool_terms) / sizeof(expr_term_fun);
}


//////////////////////////////////////////////////////////////
///
/// While the above code is used for creating random 
/// expressions, the following code is used for parsing.
///
//////////////////////////////////////////////////////////////


/// Set up a table that maps from opcode to parse function ///
genetic_fun_binding genetic_expr_bnf_default::m_parse_table[] = {
  {k_expr_add, expr_add::parse_expr},
  {k_expr_sub, expr_sub::parse_expr},
  {k_expr_mul, expr_mul::parse_expr},
  {k_expr_cmul, expr_cmul::parse_expr},
  {k_expr_div, expr_div::parse_expr},
  {k_expr_sqrt, expr_sqrt::parse_expr},
  {k_expr_tern, expr_ternary::parse_expr},
  {k_expr_lt, expr_lt::parse_expr},
  {k_expr_gt, expr_gt::parse_expr},
  {k_expr_eq, expr_eq::parse_expr},
  {k_expr_and, expr_and::parse_expr},
  {k_expr_or, expr_or::parse_expr},
  {k_expr_not, expr_not::parse_expr},
  {k_expr_dbl_const, expr_dbl_const::parse_expr},
  {k_expr_dbl_arg, expr_dbl_arg::parse_expr},
  {k_expr_bool_const, expr_bool_const::parse_expr},
  {k_expr_bool_arg, expr_bool_arg::parse_expr}
};

uint32 
genetic_expr_bnf_default::num_bindings () const
//////////////////////////////////////////////////////////////
//
//  Returns the size of the parse table.
//
//////////////////////////////////////////////////////////////
{
  return sizeof(m_parse_table)/sizeof(genetic_fun_binding);
}


genetic_expr_bnf::genetic_expr_bnf (uint32 dbls, 
				    uint32 bools) :
//////////////////////////////////////////////////////////////
//
//  Initializes the base bnf class.  This is a pure virtual
//  class from which one can derive bnfs.  bnfs for our
//  purposes are simply dispatch tables that dictate how
//  expressions are generated and parsed.
//
//////////////////////////////////////////////////////////////
  m_num_dbl_args(dbls),
  m_num_bool_args(bools)
{
}


genetic_expr_bnf_default::genetic_expr_bnf_default (uint32 dbls,
						    uint32 bools) :
//////////////////////////////////////////////////////////////
//
//  Initializes the "default" bnf class.
//
//////////////////////////////////////////////////////////////
  genetic_expr_bnf (dbls, bools)
{
}


genetic_expr_bnf_default::genetic_expr_bnf_default (ifstream& conf_file)
//////////////////////////////////////////////////////////////
//
//  Initializes the "default" bnf class.
//
//////////////////////////////////////////////////////////////
{
  finch_claim (conf_file, "Could not open configuration file.");
  
  finch_claim ((conf_file >> m_num_dbl_args), "Badly formatted configuration.");
  finch_claim ((conf_file >> m_num_bool_args), "Badly formatted configuration.");
  conf_file.close ();
}


expr_parse_fun 
genetic_expr_bnf_default::parser_binding (const char *fname)
{
  for (uint32 i = 0; i < num_bindings(); i++) {
    if (!strcmp(fname, m_parse_table[i].fun_name))
      return m_parse_table[i].binding;
  }
  return NULL;
}
