/************************************************************************
 *
 *   File: genetic_expr_bnf.cc:
 *   
 *   Author: Mark Stephenson
 *   Date: Sat Mar 22 11:25:26 2003
 *
 *   Function: This file contains classes that describe the bnf of the
 *      expression language.  We provide a "default" bnf that can
 *      be used to create expressions.  However, if you want to 
 *      extend the bnf, you can either extend the genetic_expr_bnf
 *      class, or create a subclass of it.  
 *
 *      Here are directions for extending the default bnf:
 *      
 *      1) Choose a unique opcode for the new function, and add it 
 *         to the list of opcodes in opcodes.h, opcodes.cc.
 *
 *      2) Create a class that derives from double_expr or 
 *         boolean_expr depending on what type your function returns.
 * 
 *      3) Within the new class, create a static function that 
 *         randomly creates said function.
 *
 *      4) Likewise, create a static function for parsing.
 *
 *      For steps 2, 3 and 4, see one of the classes in double_expr.h 
 *      or boolean_expr.h.
 *
 *      5) Add the static random initialization function created in 
 *         step 3 to the appropriate dispatch table.
 *
 *           +-DISPATCH TABLE -+-DESCRIPTION-----------------+
 *           | m_dbl_terms     | Returns double/is terminal. |
 *           | m_dbl_nonterms  | Returns double/is non term. |
 *           | m_bool_terms    | Returns bool/is terminal.   |
 *           | m_bool_nonterms | Returns bool/is non term.   |
 *           +-----------------+-----------------------------+
 *
 *      6) Add the static parsing function created in step 4 to the
 *         genetic_expr_bnf::parse_table.  The entry in the table is
 *         a binding from the opcode you choose in step 1 to the
 *         function you created in step 4.  Look at the table, you'll
 *         see what I'm talking about.
 *
 *      If you wish to subclass genetic_expr_bnf, you shoud know that
 *      this class is simply used a dispatch table for parsing and
 *      creaing random expressions.
 *      
 *
 ************************************************************************/

#ifndef __GENETIC_EXPR_BNF_H__
#define __GENETIC_EXPR_BNF_H__

#include "opcodes.h"
#include "common.h"
#include <fstream.h>

class genetic_expr;
class genetic_lexer;
class genetic_expr_bnf;

typedef genetic_expr* 
(*expr_nonterm_fun)(genetic_expr_bnf*,uint32,genetic_expr*);

typedef genetic_expr* 
(*expr_term_fun)(genetic_expr_bnf*,genetic_expr*);

typedef genetic_expr* 
(*expr_parse_fun)(genetic_expr_bnf*,genetic_expr*,genetic_lexer*);

typedef struct {
  char *fun_name;
  expr_parse_fun binding;
} genetic_fun_binding;


/**********************************************************************
 *
 *  This is a pure virtual class that contains information about the
 *  bnf that a genetic_expr should use.  It also contains other 
 *  global information such as the number of each type of argument.
 *  A subclass of this class should be used to assign an bnf to a
 *  "class" of expressions.  See the genetic_expr_info class for 
 *  an example.
 *
 *********************************************************************/
class genetic_expr_bnf
{
 public:

  virtual ~genetic_expr_bnf () { }

  uint32   dargs () const       { return m_num_dbl_args; }
  uint32   bargs () const       { return m_num_bool_args; }

  // Parse table/Dispatch information that needs to be
  // defined.  See the example in common.cc.
  virtual uint32 num_dbl_nonterms () const = 0;
  virtual uint32 num_dbl_terms () const  = 0;
  virtual uint32 num_bool_nonterms () const = 0;
  virtual uint32 num_bool_terms () const = 0;
  virtual uint32 num_bindings () const = 0;

  virtual expr_nonterm_fun* dbl_nonterms () = 0;
  virtual expr_term_fun* dbl_terms () = 0;
  virtual expr_nonterm_fun* bool_nonterms () = 0;
  virtual expr_term_fun* bool_terms () = 0;

  virtual expr_parse_fun parser_binding (const char *fname) = 0;

 protected:

  genetic_expr_bnf (uint32 dbls, uint32 bools);
  genetic_expr_bnf () {};

  uint32 m_num_dbl_args;  // Number of double arguments.
  uint32 m_num_bool_args; // Number of boolean arguments.
};



/**********************************************************************
 *
 *  A default bnf.  The bnf in this case is statically defined in
 *  genetic_expr_bnf.cc.  The only dynamic parameters of this bnf
 *  are the number of double arguments and the number of boolean
 *  arguments from which to choose when randomly creating new
 *  expressions. 
 *  
 **********************************************************************/
class genetic_expr_bnf_default : public genetic_expr_bnf
{
 public:

  genetic_expr_bnf_default (ifstream&);
  genetic_expr_bnf_default (uint32 dbls, uint32 bools);
  ~genetic_expr_bnf_default () {}

  virtual uint32 num_dbl_nonterms () const;
  virtual uint32 num_dbl_terms () const;
  virtual uint32 num_bool_nonterms () const;
  virtual uint32 num_bool_terms () const;
  virtual uint32 num_bindings () const;

  virtual expr_nonterm_fun* dbl_nonterms ()  { return m_dbl_nonterms; }
  virtual expr_term_fun* dbl_terms ()        { return m_dbl_terms; }
  virtual expr_nonterm_fun* bool_nonterms () { return m_bool_nonterms; }
  virtual expr_term_fun* bool_terms ()       { return m_bool_terms; }

  virtual expr_parse_fun parser_binding (const char *fname);

 public:

  static expr_nonterm_fun m_dbl_nonterms[];
  static expr_term_fun m_dbl_terms[];
  static expr_nonterm_fun m_bool_nonterms[];
  static expr_term_fun m_bool_terms[];

  static genetic_fun_binding m_parse_table[];  
};

#endif
