/****************************************************************************
 *
 *   File: compile_harness.cc:
 *   
 *   Author: Mark Stephenson
 *   Date: Thu Feb  6 18:53:03 2003
 *
 *   Function: Controls the compilation and execution of the benchmark(s)
 *     in the benchmark suite.
 *
 ***************************************************************************/

#include <stdio.h>
#include <unistd.h>
#include "benchmark_set.h"
#include "cmdparse.h"
#include "genetic_expr.h"
#include "simple_harness.h"

/*************************************************************
 *
 *  Parses the command line, then creates the initial
 *  population.  The GP algorithm is then run.
 *
 ************************************************************/
int
main (int32 argc, char **argv)
{
  bool   help;
  bool   quantize;
  bool   classifier;
  uint32 pop_size;
  uint32 num_gens;
  uint32 mutate_rate;
  uint32 mortality;
  char*  init_pop_fname;
  char*  save_pop_fname;
  char*  finchhome;
  char   rngtable[128];
  uint32 tournament_size;
  uint32 dss_set_size;
  uint32 max_genome_size;
  uint32 initial_height;
  uint32 elitist;
  uint32 sunset;
  uint32 seed;
  benchmark_set* benches;

  help = false;
  classifier = false;

  static cmd_line_option options[] = {
    {CLO_INT,   "-gens", "100", &num_gens},
    {CLO_NOARG, "-help",  "",   &help},
    {CLO_STRING,"-savepop", "", &save_pop_fname},
    {CLO_INT,   "-gensize", "180", &max_genome_size},
    {CLO_INT,   "-height", "3", &initial_height},
    {CLO_STRING,"-initial", "", &init_pop_fname},
    {CLO_INT,   "-dsssize", "6", &dss_set_size},
    {CLO_INT,   "-elitist", "1", &elitist},
    {CLO_INT,   "-mutate", "8", &mutate_rate},
    {CLO_INT,   "-mortality", "22", &mortality},
    {CLO_INT,   "-popsize", "400", &pop_size},
    {CLO_NOARG, "-quantize", "", &quantize},
    {CLO_NOARG, "-classifier", "", &classifier},
    {CLO_INT,   "-seed", "42", &seed},
    {CLO_INT,   "-sunset", "1", &sunset},
    {CLO_INT,   "-tournament", "7", &tournament_size}
  };

  for (int32 i = 0; i < argc; i++)
    cout << argv[i] << " ";
  cout << endl;

  parse_cmd_line (argc, argv, options, 
		  sizeof(options)/sizeof(cmd_line_option));

  if ((argc < 2) | help) 
  {
    cout << "\nUsage: " << argv[0];
    cout << " [options] <benchmarks_file>\n\n";
    if (help) 
    {
      cout << "\n\toptions: \n";
      cout << "\t -classifier creates classification (binary) priority funs.\n";
      cout << "\t -dsssize    dynamic subset selection size (6).\n";
      cout << "\t -elitist    specify number of top exprs to keep (1).\n";
      cout << "\t -initial    file with initial population.\n";
      cout << "\t -gens       specifies the number of generations (100).\n";
      cout << "\t -gensize    the maximum genome size (180).\n";
      cout << "\t -help       print help information.\n";
      cout << "\t -height     specifies initial expression heights (3).\n";
      cout << "\t -mutate     specifies the mutate rate (8).\n";
      cout << "\t -mortality  percent of population that dies (22).\n";
      cout << "\t -popsize    specifies the population size. (400)\n";
      cout << "\t -quantize   quantizes the results.\n";
      cout << "\t -savepop    save population after every generation.\n";
      cout << "\t -seed       seeds the random number generator (42).\n";
      cout << "\t -sunset     how many gens. to keep best expr (1).\n";
      cout << "\t -tournament specifies the tournament size (7).\n\n";
    }
    exit (1);
  }

  finch_claim (pop_size >= 10, "Unreasonable population size.");
  finch_claim (num_gens >= 1, "Invalid number of generations.");
  finch_claim (mutate_rate <= 100, "Invalid rate of mutation.");
  finch_claim (mortality <= 100, "Invalid rate of mortality.");
  finch_claim (initial_height < 9, "Set initial height lower.");
  finch_claim ((tournament_size >= 1) && (tournament_size <= pop_size),
	       "Invalid tournament size.");
  
  // Make sure that the FINCHHOME environment variable is set.
  finchhome = getenv ("FINCHHOME");
  finch_claim (finchhome != NULL, 
	       "Please set the FINCHHOME environment variable.");

  cout << "\nFinch Version 6-9-2003\n\n";
  cout << "Initial seed      : " << seed << endl;

  // Initialize the random number generators.
  srand48 (seed);
  initstate (seed, rngtable, sizeof(rngtable));

  // Create the benchmarks class.
  benches = new benchmark_set (argv[1]);

  // Create the GP harness.
  simple_harness harness (pop_size,
			  num_gens,
			  (double) ((double)mutate_rate / 100.0),
			  (double) ((double)mortality / 100.0),
			  init_pop_fname,
			  save_pop_fname,
			  finchhome,
			  tournament_size,
			  dss_set_size,
			  max_genome_size,
			  initial_height,
			  elitist,
			  sunset,
			  quantize,
			  classifier,
			  benches);

  // Run the GP system.
  harness.search_for_best ();
}



