/************************************************************************
 *
 *   File: simple_harness.h:
 *   
 *   Author: Mark Stephenson
 *   Date: Wed Aug 21 19:52:38 2002
 *
 *   Function: This harness uses the file system to communicate
 *     expressions and fitnesses.  It's called simple_harness because
 *     it is very simple.  It uses a directory (set according to an
 *     environment variable) to communicate all information via files.
 *
 ***********************************************************************/

#ifndef __SIMPLE_HARNESS_H__
#define __SIMPLE_HARNESS_H__


#include "genetic_expr.h"
#include "benchmark_set.h"
#include <vector>
#include <set>
#include <iostream.h>
#include <functional>
#include <math.h>
#include <bits/nan.h>


class simple_harness
{
 public:

  // Create a new harness.
  simple_harness (uint32  population_size,
		  uint32  generations,
		  double  mutate_rate,
		  double  mortality,
		  char*   initial_pop_fname,
		  char*   save_pop_fname,
		  char*   finch_home,
		  uint32  tournament_size,
		  uint32  dss_set_size,
		  uint32  maximum_genome_size,
		  uint32  initial_height,
		  uint32  elitist,
		  uint32  sunset,
		  bool    quantize_results,
		  bool    classifier,
		  benchmark_set *benches);

  
  // Kill the harness.
  ~simple_harness ();

  // Searches for the best expression to the problem.
  genetic_expr *search_for_best ();

 protected:

  // Creates the configuration file.
  void create_configuration ();

  // Creates the initial population file.
  void create_initial_population (uint32, uint32, char*);

  // Clears the work directories.
  void clear_evaluate_dir ();

  // Compiles and runs the benchmarks with the given expression.
  void dispatch (benchmark_info*, genetic_expr*);

  // Executes the commands in the command list.
  void execute_commands (command_list *cmds, bool = false);

  //  function that farms out the jobs and collects results.
  void farm_out_jobs ();

  // Performs crossover and mutation.
  void crossover_and_mutate ();
  
  // Prints the population.
  void print_population (ostream&, uint32, bool = false) const;
  
  // Prepares for next generation.
  void prepare_for_generation ();
  
  // Splits fitnesses into quanta.
  void quantize_population ();

  uint32 popsize () const   { return m_population.size(); }
  
 protected:

  string  m_workdir;
  uint32  m_memoized;
  uint32  m_gens;
  double  m_mutate_rate;
  double  m_mortality;
  char*   m_save_pop_fname;

  ofstream *m_state_file;

  uint32  m_tournament_size;
  uint32  m_dss_set_size;
  uint32  m_max_genome_size;
  uint32  m_elitist;
  uint32  m_sunset;

  benchmark_set *m_benches;

  bool m_quantize_results;
  bool m_classifier;
  genetic_expr_bnf *m_bnf;
  set<double> *m_quanta;

  vector<genetic_expr*> m_population;
};


struct expr_lt : public binary_function<genetic_expr*,genetic_expr*,bool>
{
  // Returns true if x is `less fit' than y (i.e., x has a 
  // higher speedup number).
  bool operator () (genetic_expr *x, genetic_expr *y)
  {
    double fity = y->get_fitness ();
    if (isnan(fity)) return true;
    double fitx = x->get_fitness ();
    if (isnan(fitx)) return false;
    
    // Enforce parsimony.
    if (fitx == fity)
      return (x->size() < y->size());
    else
      return (fitx < fity);
  }
};


#endif
